"""
PyPatchwork - Python Library for Patchwork REST API

Copyright (C) 2023  Tedd Ho-Jeong An <hj.tedd.an@gmail.com>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/
"""
from patchwork.Patchwork import Patchwork

from .PatchworkException import (
    PatchworkException,
    BadRequestException,
    ForbiddenException,
    NotFoundException,
    ConflictException,
)



"""
PyPatchwork - Python Library for Patchwork REST API

Copyright (C) 2023  Tedd Ho-Jeong An <hj.tedd.an@gmail.com>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/
"""
import unittest
import patchwork


class TestPatch(unittest.TestCase):
    def setUp(self):
        self.pw = patchwork.Patchwork("https://patchwork.kernel.org")
        self.patch = self.pw.get_patch(12567487)

    def testBasicAttributes(self):
        self.assertEqual(self.patch.id, 12567487)
        self.assertEqual(
            self.patch.url, "https://patchwork.kernel.org/api/patches/12567487/"
        )
        self.assertEqual(
            self.patch.web_url,
            "https://patchwork.kernel.org/project/bluetooth/patch/20211018172833.534191-2-hj.tedd.an@gmail.com/",
        )
        self.assertEqual(
            self.patch.msgid, "<20211018172833.534191-2-hj.tedd.an@gmail.com>"
        )
        self.assertEqual(
            self.patch.list_archive_url,
            "https://lore.kernel.org/r/20211018172833.534191-2-hj.tedd.an@gmail.com",
        )
        self.assertEqual(self.patch.date, "2021-10-18T17:28:25")
        self.assertEqual(
            self.patch.name, "[BlueZ,1/9] device: Fix unchecked return value"
        )
        self.assertEqual(self.patch.commit_ref, None)
        self.assertEqual(self.patch.pull_url, None)
        self.assertEqual(self.patch.state, "accepted")
        self.assertEqual(self.patch.archived, False)
        self.assertEqual(self.patch.hash, "f8ffe4bbc6a3f3f2ebf2da58a6d10348f8e4903d")
        self.assertEqual(
            self.patch.mbox,
            "https://patchwork.kernel.org/project/bluetooth/patch/20211018172833.534191-2-hj.tedd.an@gmail.com/mbox/",
        )
        self.assertEqual(
            self.patch.comments,
            "https://patchwork.kernel.org/api/patches/12567487/comments/",
        )
        self.assertEqual(self.patch.check, "success")
        self.assertEqual(
            self.patch.checks,
            "https://patchwork.kernel.org/api/patches/12567487/checks/",
        )
        self.assertEqual(self.patch.tags, {})
        self.assertEqual(self.patch.related, [])
        self.assertEqual(self.patch.prefixes, ["BlueZ", "1/9"])

    def testProject(self):
        project = self.patch.get_project()
        self.assertEqual(project.id, 395)
        self.assertEqual(project.url, "https://patchwork.kernel.org/api/projects/395/")
        self.assertEqual(project.name, "Bluetooth")
        self.assertEqual(project.link_name, "bluetooth")
        self.assertEqual(project.list_id, "linux-bluetooth.vger.kernel.org")
        self.assertEqual(project.list_email, "linux-bluetooth@vger.kernel.org")
        self.assertEqual(project.web_url, "")
        self.assertEqual(project.scm_url, "")
        self.assertEqual(project.webscm_url, "")
        self.assertEqual(project.list_archive_url, "")
        self.assertEqual(
            project.list_archive_url_format, "https://lore.kernel.org/r/{}"
        )
        self.assertEqual(project.commit_url_format, "")
        self.assertTrue(self.patch.content.startswith("From: Tedd Ho-Jeong An"))

    def testSubmitter(self):
        submitter = self.patch.get_submitter()
        self.assertEqual(submitter.id, 196023)
        self.assertEqual(
            submitter.url, "https://patchwork.kernel.org/api/people/196023/"
        )
        self.assertEqual(submitter.name, "Tedd An")
        self.assertEqual(submitter.email, "hj.tedd.an@gmail.com")

    def testDelegate(self):
        delegate = self.patch.get_delegate()
        self.assertEqual(delegate.id, 103487)
        self.assertEqual(delegate.url, "https://patchwork.kernel.org/api/users/103487/")
        self.assertEqual(delegate.username, "vudentz")
        self.assertEqual(delegate.first_name, "Luiz")
        self.assertEqual(delegate.last_name, "Von Dentz")
        self.assertEqual(delegate.email, "luiz.dentz@gmail.com")

    def testSeries(self):
        series_list = self.patch.get_series()
        self.assertIsNotNone(series_list)
        self.assertTrue(len(series_list) >= 1)
        for series in series_list:
            self.assertIsInstance(series, patchwork.Series.Series)
        series_0 = series_list[0]
        self.assertEqual(series_0.id, 565705)
        self.assertEqual(
            series_0.url, "https://patchwork.kernel.org/api/series/565705/"
        )
        self.assertEqual(
            series_0.web_url,
            "https://patchwork.kernel.org/project/bluetooth/list/?series=565705",
        )
        self.assertEqual(series_0.date, "2021-10-18T17:28:24")
        self.assertEqual(series_0.name, "Fix unchecked return value")
        self.assertEqual(series_0.version, 1)
        self.assertEqual(
            series_0.mbox, "https://patchwork.kernel.org/series/565705/mbox/"
        )

    def testHeaders(self):
        headers = self.patch.headers
        # Only check a few of them..
        self.assertIn("From", headers)
        self.assertEqual(headers["From"], "Tedd Ho-Jeong An <hj.tedd.an@gmail.com>")
        self.assertIn("Subject", headers)
        self.assertEqual(
            headers["Subject"], "[BlueZ PATCH 1/9] device: Fix unchecked return value"
        )
        self.assertIn("Message-Id", headers)
        self.assertEqual(
            headers["Message-Id"], "<20211018172833.534191-2-hj.tedd.an@gmail.com>"
        )



"""
PyPatchwork - Python Library for Patchwork REST API

Copyright (C) 2023  Tedd Ho-Jeong An <hj.tedd.an@gmail.com>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/
"""
import patchwork.Bundle
import patchwork.Connection
import patchwork.Cover
import patchwork.Pagination
import patchwork.Patch
import patchwork.People
import patchwork.Project
import patchwork.Series
import patchwork.User


class Patchwork:
    """
    This is the main class to access the Patchwork API.
    """

    def __init__(
        self,
        base_url,
        token=None,
    ):
        """
        :param base_url: Base url of the Patchwork site
        :type base_url: string
        """
        assert isinstance(base_url, str), "Invalid parameter: base_url"
        if token is not None:
            assert isinstance(token, str), "Invalid parameter: token"

        # Setup connection object
        self.__connection = patchwork.Connection.Connection(
            base_url,
            token,
        )

    def get_user(self, id):
        """
        Show a user

        :calls: GET /api/users/{id}/
        :param id: A unique integer value identifying this user
        :type id: integer
        :rtype: :class:`patchwork.User.User`
        """
        assert isinstance(id, int)

        headers, data = self.__connection.request(
            "GET",
            f"/api/users/{id}",
        )
        return patchwork.User.User(self.__connection, data)

    def get_all_users(self):
        """
        List Users

        :calls: GET /api/users/
        :rtype: :class:`patchwork.Pagination.Pagination` of :class:`patchwork.User.User`
        """
        return patchwork.Pagination.Pagination(
            patchwork.User.User,
            self.__connection,
            "/api/users",
        )

    def get_project(self, id):
        """
        Show a project

        :calls: GET /api/projects/{id}
        :param id: A unique integer value identifying this project
        :type id: integer
        :rtype: :class:`patchwork.Project.Project`
        """
        assert isinstance(id, int)

        headers, data = self.__connection.request(
            "GET",
            f"/api/projects/{id}",
        )
        return patchwork.Project.Project(self.__connection, data)

    def get_all_projects(self):
        """
        List projects

        :calls: GET /api/projects/
        :rtype: :class:`patchwork.Pagination.Pagination` of :class:`patchwork.Project.Project`
        """
        return patchwork.Pagination.Pagination(
            patchwork.Project.Project,
            self.__connection,
            "/api/projects",
        )

    def get_people(self, id):
        """
        Show a people

        :calls: GET /api/people/{id}
        :param id: A unique integer value identifying this person
        :type id: integer
        :rtype: :class:`patchwork.People.People`
        """
        assert isinstance(id, int)

        headers, data = self.__connection.request(
            "GET",
            f"/api/people/{id}",
        )
        return patchwork.People.People(self.__connection, data)

    def get_all_people(self):
        """
        List People

        :calls: GET /api/people/
        :rtype: :class:`patchwork.Pagination.Pagination` of :class:`patchwork.People.People`
        """
        return patchwork.Pagination.Pagination(
            patchwork.People.People,
            self.__connection,
            "/api/people",
        )

    def get_series(self, id):
        """
        Show a series

        :calls: GET /api/series/{id}
        :param id: A unique integer value identifying this series
        :type id: integer
        :rtype: :class:`patchwork.Series.Series`
        """
        assert isinstance(id, int)

        headers, data = self.__connection.request(
            "GET",
            f"/api/series/{id}",
        )
        return patchwork.Series.Series(self.__connection, data)

    def get_all_series(self):
        """
        List series

        :calls: GET /api/series/
        :rtype: :class:`patchwork.Pagination.Pagination` of :class:`patchwork.Series.Series`
        """
        return patchwork.Pagination.Pagination(
            patchwork.Series.Series,
            self.__connection,
            "/api/series",
        )

    def get_cover_letter(self, id):
        """
        Show a cover letter

        :calls: GET /api/covers/{id}
        :param id: A unique integer value identifying this cover letter
        :type id: integer
        :rtype: :class:`patchwork.Cover.Cover`
        """
        assert isinstance(id, int)

        headers, data = self.__connection.request(
            "GET",
            f"/api/covers/{id}",
        )
        return patchwork.Cover.Cover(self.__connection, data)

    def get_all_cover_letters(self):
        """
        List Cover letters

        :calls: GET /api/covers/
        :rtype: :class:`patchwork.Pagination.Pagination` of :class:`patchwork.Cover.Cover`
        """
        return patchwork.Pagination.Pagination(
            patchwork.Cover.Cover,
            self.__connection,
            "/api/covers",
        )

    def get_patch(self, id):
        """
        Show a patch

        :calls: GET /api/patches/{id}
        :param id: A unique integer value identifying this patch
        :type id: integer
        :rtype: :class:`patchwork.Patch.Patch`
        """
        assert isinstance(id, int)

        headers, data = self.__connection.request(
            "GET",
            f"/api/patches/{id}",
        )
        return patchwork.Patch.Patch(self.__connection, data)

    def get_all_patches(self):
        """
        List Patches

        :calls: GET /api/patches/
        :rtype: :class:`patchwork.Pagination.Pagination` of :class:`patchwork.Patch.Patch`
        """
        return patchwork.Pagination.Pagination(
            patchwork.Patch.Patch,
            self.__connection,
            "/api/patches",
        )

    def get_bundle(self, id):
        """
        Show a bundle

        :calls: GET /api/bundles/{id}
        :param id: A unique integer value identifying this bundle
        :type id: integer
        :rtype: :class:`patchwork.Bundle.Bundle`
        """
        assert isinstance(id, int)

        headers, data = self.__connection.request(
            "GET",
            f"/api/bundles/{id}/",
        )
        return patchwork.Bundle.Bundle(self.__connection, data)

    def get_all_bundles(self):
        """
        List Bundles

        :calls: GET /api/bundles
        :rtype: :class:`patchwork.Pagination.Pagination` of :class:`patchwork.Bundle.Bundle`
        """
        return patchwork.Pagination.Pagination(
            patchwork.Bundle.Bundle,
            self.__connection,
            "/api/bundles",
        )



"""
PyPatchwork - Python Library for Patchwork REST API

Copyright (C) 2023  Tedd Ho-Jeong An <hj.tedd.an@gmail.com>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/
"""
import json
import requests

from urllib.parse import urlparse, urlencode

from . import PatchworkException


class Connection:
    """
    Connection class
    """

    def __init__(
        self,
        base_url,
        token,
    ):
        """
        :param base_url: Base url to connect
        :type base_url: string
        :param token: Authentication token
        :type token: string
        :param per_page: Number of item per page
        :type per_page: integer
        """

        self.__base_url = base_url
        u = urlparse(base_url)
        self.__scheme = u.scheme
        self.__hostname = u.hostname
        self.__path = u.path

        if u.port:
            self.__port = u.port
        else:
            if self.__scheme == "https":
                self.__port = 443
            else:
                self.__port = 80

        self.__token = token

    def __update_auth_token(self, headers):
        if self.__token is not None:
            headers["Authorization"] = f"Token {self.__token}"

    def __make_abs_url(self, url):
        if url.startswith("http"):
            return url
        # Relative path
        return f"{self.__scheme}://{self.__hostname}{self.__path}{url}"

    def __update_url_with_params(self, url, params):
        if len(params) == 0:
            return url
        return f"{url}?{urlencode(params)}"

    def __raise_exception(self, status, headers, output):
        # 400
        if status == 400:
            cls = PatchworkException.BadRequestException
        elif status == 403:
            cls = PatchworkException.ForbiddenException
        elif status == 404:
            cls = PatchworkException.NotFoundException
        elif status == 409:
            cls = PatchworkException.ConflictException
        else:
            cls = PatchworkException.PatchworkException
        return cls(status, headers, output)

    def __request(self, method, url, parameters, headers, input):
        assert method in ["GET", "POST", "PATCH", "PUT"]
        if parameters is None:
            parameters = dict()
        if headers is None:
            headers = dict()

        # Update Authorization token to headers
        self.__update_auth_token(headers)

        # Update URL to full format
        abs_url = self.__make_abs_url(url)
        abs_url = self.__update_url_with_params(abs_url, parameters)

        print(f"ABS_URL={abs_url}")

        enc_input = None
        if input is not None:
            headers["Content-Type"] = "application/json"
            enc_input = json.dumps(input)

        resp = requests.request(method, abs_url, headers=headers, data=enc_input)
        resp_status = resp.status_code
        resp_headers = resp.headers
        if len(resp.text) == 0:
            resp_data = None
        else:
            resp_data = json.loads(resp.text)

        print(f"resp status: {resp_status}")

        # Check resp status and raise the exception
        if resp_status >= 400:
            raise self.__raise_exception(resp_status, resp_headers, resp_data)
        return resp_headers, resp_data

    def request(self, method, url, parameters=None, headers=None, input=None):
        """Send request to URL and returns header and output data in json format

        :param method: Method of the new Request object. i.e. 'GET', 'PUT'
        :type method: string
        :param url: URL for the new Request object. i.e. 'patches'/{id}'
        :type url: string
        :param parameters: Parameters for URL request, defaults to None
        :type parameters: dict, optional
        :param headers: Extra headers for URL request, defaults to None
        :type headers: dict, optional
        :param input: Input data for URL request, defaults to None
        :type input: dict, optional
        """
        return self.__request(method, url, parameters, headers, input)



"""
PyPatchwork - Python Library for Patchwork REST API

Copyright (C) 2023  Tedd Ho-Jeong An <hj.tedd.an@gmail.com>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/
"""
import patchwork.Check
import patchwork.Comment
import patchwork.Connection
import patchwork.Patch
import patchwork.People
import patchwork.Project
import patchwork.Series
import patchwork.User


class Patch:
    """
    Class for Patch object
    """

    def __init__(self, connection, attributes):
        self._connection = connection
        self._id = None
        self._url = None
        self._web_url = None
        self._project = None
        self._msgid = None
        self._list_archive_url = None
        self._date = None
        self._name = None
        self._commit_ref = None
        self._pull_url = None
        self._state = None
        self._archived = None
        self._hash = None
        self._submitter = None
        self._delegate = None
        self._mbox = None
        self._series = None
        self._comments = None
        self._check = None
        self._checks = None
        self._tags = None
        self._related = None
        self._headers = None
        self._content = None
        self._diff = None
        self._prefixes = None
        self.__update_attributes(attributes)

    def __update_attributes(self, attributes):
        if "id" in attributes:
            self._id = attributes["id"]
        if "url" in attributes:
            self._url = attributes["url"]
        if "web_url" in attributes:
            self._web_url = attributes["web_url"]
        if "project" in attributes:
            self._project = attributes["project"]
        if "msgid" in attributes:
            self._msgid = attributes["msgid"]
        if "list_archive_url" in attributes:
            self._list_archive_url = attributes["list_archive_url"]
        if "date" in attributes:
            self._date = attributes["date"]
        if "name" in attributes:
            self._name = attributes["name"]
        if "commit_ref" in attributes:
            self._commit_ref = attributes["commit_ref"]
        if "pull_url" in attributes:
            self._pull_url = attributes["pull_url"]
        if "state" in attributes:
            self._state = attributes["state"]
        if "archived" in attributes:
            self._archived = attributes["archived"]
        if "hash" in attributes:
            self._hash = attributes["hash"]
        if "submitter" in attributes:
            self._submitter = attributes["submitter"]
        if "delegate" in attributes:
            self._delegate = attributes["delegate"]
        if "mbox" in attributes:
            self._mbox = attributes["mbox"]
        if "series" in attributes:
            self._series = attributes["series"]
        if "comments" in attributes:
            self._comments = attributes["comments"]
        if "check" in attributes:
            self._check = attributes["check"]
        if "checks" in attributes:
            self._checks = attributes["checks"]
        if "tags" in attributes:
            self._tags = attributes["tags"]
        if "related" in attributes:
            self._related = attributes["related"]
        if "headers" in attributes:
            self._headers = attributes["headers"]
        if "content" in attributes:
            self._content = attributes["content"]
        if "diff" in attributes:
            self._diff = attributes["diff"]
        if "prefixes" in attributes:
            self._prefixes = attributes["prefixes"]

    @property
    def id(self):
        """
        :type: integer
        """
        return self._id

    @property
    def url(self):
        """
        :type: string
        """
        return self._url

    @property
    def web_url(self):
        """
        :type: string
        """
        return self._web_url

    @property
    def msgid(self):
        """
        :type: string
        """
        return self._msgid

    @property
    def list_archive_url(self):
        """
        :type: string
        """
        return self._list_archive_url

    @property
    def date(self):
        """
        :type: string
        """
        return self._date

    @property
    def name(self):
        """
        :type: string
        """
        return self._name

    @property
    def commit_ref(self):
        """
        :type: integer
        """
        return self._commit_ref

    @property
    def pull_url(self):
        """
        :type: boolean
        """
        return self._pull_url

    @property
    def state(self):
        """
        :type: string
        """
        return self._state

    @property
    def archived(self):
        """
        :type: string
        """
        return self._archived

    @property
    def hash(self):
        """
        :type: string
        """
        return self._hash

    @property
    def mbox(self):
        """
        :type: string
        """
        return self._mbox

    @property
    def comments(self):
        """
        :type: string
        """
        return self._comments

    @property
    def check(self):
        """
        :type: string
        """
        return self._check

    @property
    def checks(self):
        """
        :type: string
        """
        return self._checks

    @property
    def tags(self):
        """
        :type: dict
        """
        return self._tags

    @property
    def headers(self):
        """
        :type: dict
        """
        return self._headers

    @property
    def content(self):
        """
        :type: string
        """
        return self._content

    @property
    def diff(self):
        """
        :type: string
        """
        return self._diff

    @property
    def prefixes(self):
        """
        :type: list
        """
        return self._prefixes

    @property
    def related(self):
        return self._related

    def get_project(self):
        """
        Project detail (partial)

        :rtype: :class:`patchwork.Project.Project`
        """
        return patchwork.Project.Project(self._connection, self._project)

    def get_submitter(self):
        """
        Submitter detail (partial)

        :rtype: :class:`patchwork.People.People`
        """
        return patchwork.People.People(self._connection, self._submitter)

    def get_series(self):
        """
        Series detail (partial)

        :rtype: List of :class:`patchwork.Series.Series`
        """
        series = []
        for ser in self._series:
            series.append(patchwork.Series.Series(self._connection, ser))
        return series

    def get_related(self):
        """
        List of related patch detail (partial)

        :rtype: List of :class:`patchwork.Patch.Patch`
        """
        related = []
        for patch in self._related:
            related.append(patchwork.Patch.Patch(self._connection, patch))
        return related

    def get_delegate(self):
        """
        Delegated User detail (partial)

        :rtype: :class:`patchwork.User.User`
        """
        return patchwork.User.User(self._connection, self._delegate)

    def get_comments(self):
        """
        List comments

        :calls: GET /api/patches/{id}/comments
        :rtype: List of :class:`patchwork.Comment.Comment`
        """
        headers, data = self._connection.request(
            "GET", f"/api/patches/{self._id}/comments"
        )
        comments = []
        for comment in data:
            comments.append(patchwork.Comment.Comment(self._connection, comment))
        return comments

    def get_checks(self):
        """
        List checks

        :calls: GET /api/patches/{id}/checks
        :rtype: List of :class:`patchwork.Check.Check`
        """
        headers, data = self._connection.request(
            "GET", f"/api/patches/{self._id}/checks"
        )
        checks = []
        for check in data:
            checks.append(patchwork.Check.Check(self._connection, check))
        return checks

    def get_check(self, check_id):
        """
        Show a check

        :calls: GET /api/patches/{id}/checks/{check_id}
        :param check_id: A unique integer value identifying this check
        :type check_id: integer
        :rtype: :class:`patchwork.Check.Check`
        """
        headers, data = self._connection.request(
            "GET", f"/api/patches/{self._id}/checks/{check_id}"
        )
        return patchwork.Check.Check(self._connection, data)



"""
PyPatchwork - Python Library for Patchwork REST API

Copyright (C) 2023  Tedd Ho-Jeong An <hj.tedd.an@gmail.com>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/
"""
import unittest
import patchwork


class TestSeries(unittest.TestCase):
    def setUp(self):
        self.pw = patchwork.Patchwork("https://patchwork.kernel.org")
        self.series = self.pw.get_series(565705)

    def testBasicAttributes(self):
        self.assertEqual(self.series.id, 565705)
        self.assertEqual(
            self.series.url, "https://patchwork.kernel.org/api/series/565705/"
        )
        self.assertEqual(
            self.series.web_url,
            "https://patchwork.kernel.org/project/bluetooth/list/?series=565705",
        )
        self.assertEqual(self.series.name, "Fix unchecked return value")
        self.assertEqual(self.series.date, "2021-10-18T17:28:24")
        self.assertEqual(self.series.version, 1)
        self.assertEqual(self.series.total, 9)
        self.assertEqual(self.series.received_total, 9)
        self.assertTrue(self.series.received_all)
        self.assertEqual(
            self.series.mbox, "https://patchwork.kernel.org/series/565705/mbox/"
        )
        self.assertIsInstance(self.series.get_project(), patchwork.Project.Project)
        self.assertIsInstance(self.series.get_submitter(), patchwork.People.People)
        self.assertIsInstance(self.series.get_cover_letter(), patchwork.Cover.Cover)
        patches = self.series.get_patches()
        self.assertIsNotNone(patches)
        for patch in patches:
            self.assertIsInstance(patch, patchwork.Patch.Patch)

    def testProject(self):
        project = self.series.get_project()
        self.assertEqual(project.id, 395)
        self.assertEqual(project.url, "https://patchwork.kernel.org/api/projects/395/")
        self.assertEqual(project.name, "Bluetooth")
        self.assertEqual(project.link_name, "bluetooth")
        self.assertEqual(project.list_id, "linux-bluetooth.vger.kernel.org")
        self.assertEqual(project.list_email, "linux-bluetooth@vger.kernel.org")
        self.assertEqual(project.web_url, "")
        self.assertEqual(project.scm_url, "")
        self.assertEqual(project.webscm_url, "")
        self.assertEqual(project.list_archive_url, "")
        self.assertEqual(
            project.list_archive_url_format, "https://lore.kernel.org/r/{}"
        )
        self.assertEqual(project.commit_url_format, "")

    def testSubmitter(self):
        submitter = self.series.get_submitter()
        self.assertEqual(submitter.id, 196023)
        self.assertEqual(
            submitter.url, "https://patchwork.kernel.org/api/people/196023/"
        )
        self.assertEqual(submitter.name, "Tedd An")
        self.assertEqual(submitter.email, "hj.tedd.an@gmail.com")

    def testCoverLetter(self):
        cover = self.series.get_cover_letter()
        self.assertEqual(cover.id, 12567485)
        self.assertEqual(cover.url, "https://patchwork.kernel.org/api/covers/12567485/")
        self.assertEqual(
            cover.web_url,
            "https://patchwork.kernel.org/project/bluetooth/cover/20211018172833.534191-1-hj.tedd.an@gmail.com/",
        )
        self.assertEqual(cover.msgid, "<20211018172833.534191-1-hj.tedd.an@gmail.com>")
        self.assertEqual(
            cover.list_archive_url,
            "https://lore.kernel.org/r/20211018172833.534191-1-hj.tedd.an@gmail.com",
        )
        self.assertEqual(cover.date, "2021-10-18T17:28:24")
        self.assertEqual(cover.name, "[BlueZ,0/9] Fix unchecked return value")
        self.assertEqual(
            cover.mbox,
            "https://patchwork.kernel.org/project/bluetooth/cover/20211018172833.534191-1-hj.tedd.an@gmail.com/mbox/",
        )

    def testPatches(self):
        patches = self.series.get_patches()
        patch_0 = patches[0]
        self.assertEqual(patch_0.id, 12567487)
        self.assertEqual(
            patch_0.url, "https://patchwork.kernel.org/api/patches/12567487/"
        )
        self.assertEqual(
            patch_0.web_url,
            "https://patchwork.kernel.org/project/bluetooth/patch/20211018172833.534191-2-hj.tedd.an@gmail.com/",
        )
        self.assertEqual(
            patch_0.msgid, "<20211018172833.534191-2-hj.tedd.an@gmail.com>"
        )
        self.assertEqual(
            patch_0.list_archive_url,
            "https://lore.kernel.org/r/20211018172833.534191-2-hj.tedd.an@gmail.com",
        )
        self.assertEqual(patch_0.date, "2021-10-18T17:28:25")
        self.assertEqual(patch_0.name, "[BlueZ,1/9] device: Fix unchecked return value")
        self.assertEqual(
            patch_0.mbox,
            "https://patchwork.kernel.org/project/bluetooth/patch/20211018172833.534191-2-hj.tedd.an@gmail.com/mbox/",
        )



"""
PyPatchwork - Python Library for Patchwork REST API

Copyright (C) 2023  Tedd Ho-Jeong An <hj.tedd.an@gmail.com>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/
"""


class UserSettings:
    """
    Class for User setting object
    """

    def __init__(self, connection, attributes):
        self._connection = connection
        self._send_email = None
        self._items_per_page = None
        self._show_ids = None
        # Update attributes
        self._update_attributes(attributes)

    def _update_attributes(self, attributes):
        if "send_email" in attributes:
            self._send_email = attributes["send_email"]
        if "items_per_page" in attributes:
            self._items_per_page = attributes["items_per_page"]
        if "show_ids" in attributes:
            self._show_ids = attributes["show_ids"]

    def __repr__(self):
        output = ""
        if self._send_email is not None:
            output += f"{self._send_email}"
        if self._items_per_page is not None:
            output += f"{self._items_per_page}"
        if self._show_ids is not None:
            output += f"{self._show_ids}"
        return output

    @property
    def send_email(self):
        """
        :type: boolean
        """
        return self._send_email

    @property
    def items_per_page(self):
        """
        :type: integer
        """
        return self._items_per_page

    @property
    def show_ids(self):
        """
        :type: boolean
        """
        return self._show_ids



"""
PyPatchwork - Python Library for Patchwork REST API

Copyright (C) 2023  Tedd Ho-Jeong An <hj.tedd.an@gmail.com>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/
"""
import patchwork.UserSettings


class User:
    """
    Class for User object
    """

    def __init__(self, connection, attributes):
        self._connection = connection
        self._id = None
        self._url = None
        self._username = None
        self._first_name = None
        self._last_name = None
        self._email = None
        self._settings = None
        # Update attributes
        self._update_attributes(attributes)

    def _update_attributes(self, attributes):
        if attributes is None:
            return
        if "id" in attributes:
            self._id = attributes["id"]
        if "url" in attributes:
            self._url = attributes["url"]
        if "username" in attributes:
            self._username = attributes["username"]
        if "first_name" in attributes:
            self._first_name = attributes["first_name"]
        if "last_name" in attributes:
            self._last_name = attributes["last_name"]
        if "email" in attributes:
            self._email = attributes["email"]
        if "settings" in attributes:
            self._settings = attributes["settings"]

    def __repr__(self):
        return f"User(id:{self._id} username:{self._username})"

    @property
    def id(self):
        """
        :type: integer
        """
        return self._id

    @property
    def url(self):
        """
        :type: string
        """
        return self._url

    @property
    def username(self):
        """
        :type: string
        """
        return self._username

    @property
    def first_name(self):
        """
        :type: string
        """
        return self._first_name

    @property
    def last_name(self):
        """
        :type: string
        """
        return self._last_name

    @property
    def email(self):
        """
        :type: string
        """
        return self._email

    @property
    def settings(self):
        """
        :type: :class:`patchwork.UserSettings.UserSettings`
        """
        return patchwork.UserSettings.UserSettings(self._connection, self._settings)



"""
PyPatchwork - Python Library for Patchwork REST API

Copyright (C) 2023  Tedd Ho-Jeong An <hj.tedd.an@gmail.com>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/
"""
import patchwork.Patch
import patchwork.Project
import patchwork.User


class Bundle:
    """
    Class for Bundle object
    """

    def __init__(self, connection, attributes):
        self._connection = connection
        self._id = None
        self._url = None
        self._web_url = None
        self._project = None
        self._name = None
        self._owner = None
        self._patches = None
        self._public = None
        self._mbox = None
        self.__update_attributes(attributes)

    def __update_attributes(self, attributes):
        if "id" in attributes:
            self._id = attributes["id"]
        if "url" in attributes:
            self._url = attributes["url"]
        if "web_url" in attributes:
            self._web_url = attributes["web_url"]
        if "project" in attributes:
            self._project = attributes["project"]
        if "name" in attributes:
            self._name = attributes["name"]
        if "owner" in attributes:
            self._owner = attributes["owner"]
        if "patches" in attributes:
            self._patches = attributes["patches"]
        if "public" in attributes:
            self._public = attributes["public"]
        if "mbox" in attributes:
            self._mbox = attributes["mbox"]

    @property
    def id(self):
        """
        :type: integer
        """
        return self._id

    @property
    def url(self):
        """
        :type: string
        """
        return self._url

    @property
    def web_url(self):
        """
        :type: string
        """
        return self._web_url

    @property
    def name(self):
        """
        :type: string
        """
        return self._name

    @property
    def public(self):
        """
        :type: string
        """
        return self._public

    @property
    def mbox(self):
        """
        :type: string
        """
        return self._mbox

    def get_project(self):
        """
        Project detail (partial)

        :rtype: :class:`patchwork.Project.Project`
        """
        return patchwork.Project.Project(self._connection, self._project)

    def get_owner(self):
        """
        Owner detail (partial)

        :rtype: :class:`patchwork.User.User`
        """
        return patchwork.User.User(self._connection, self._owner)

    def get_patches(self):
        """
        List of Patches

        :rtype: List of :class:`patchwork.Patch.Patch`
        """
        patches = []
        for patch in self._patches:
            patches.append(patchwork.Patch.Patch(self._connection, patch))
        return patches



"""
PyPatchwork - Python Library for Patchwork REST API

Copyright (C) 2023  Tedd Ho-Jeong An <hj.tedd.an@gmail.com>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/
"""


class PatchworkException(Exception):
    """
    Exception class for Patchwork
    """

    def __init__(self, status, headers, data):
        super().__init__()
        self.__status = status
        self.__headers = headers
        self.__data = data
        self.__detail = self.__get_detail()

    def __get_detail(self):
        if "Content-Type" in self.__headers:
            if self.__headers["Content-Type"] == "application/json":
                if "detail" in self.__data:
                    return self.__data["detail"]
        return "Generic Exception"

    @property
    def status(self):
        """
        HTTP status
        """
        return self.__status

    @property
    def detail(self):
        """
        Exception detail provided by the Patchwork server, if available
        """
        return self.__detail

    def __str__(self):
        return f"{self.__status} {self.__detail}"


class BadRequestException(PatchworkException):
    """
    400 Bad Request - Invalid Request
    """


class ForbiddenException(PatchworkException):
    """
    403 Forbidden - Fobidden
    """


class NotFoundException(PatchworkException):
    """
    404 Not Found - Not Found
    """


class ConflictException(PatchworkException):
    """
    409 Conflict - Conflict
    """



"""
PyPatchwork - Python Library for Patchwork REST API

Copyright (C) 2023  Tedd Ho-Jeong An <hj.tedd.an@gmail.com>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/
"""
from .User import User


class Check:
    """
    Class for Chceck object
    """

    def __init__(self, connection, attributes):
        self._connection = connection
        self._id = None
        self._url = None
        self._user = None
        self._date = None
        self._state = None
        self._target_url = None
        self._context = None
        self._description = None
        self.__update_attributes(attributes)

    def __update_attributes(self, attributes):
        if "id" in attributes:
            self._id = attributes["id"]
        if "url" in attributes:
            self._url = attributes["url"]
        if "user" in attributes:
            self._user = attributes["user"]
        if "date" in attributes:
            self._date = attributes["date"]
        if "state" in attributes:
            self._state = attributes["state"]
        if "target_url" in attributes:
            self._target_url = attributes["target_url"]
        if "context" in attributes:
            self._context = attributes["context"]
        if "description" in attributes:
            self._description = attributes["description"]

    @property
    def id(self):
        """
        :type: integer
        """
        return self._id

    @property
    def url(self):
        """
        :type: string
        """
        return self._url

    @property
    def date(self):
        """
        :type: string
        """
        return self._date

    @property
    def state(self):
        """
        :type: string
        """
        return self._state

    @property
    def target_url(self):
        """
        :type: string
        """
        return self._target_url

    @property
    def context(self):
        """
        :type: string
        """
        return self._context

    @property
    def description(self):
        """
        :type: dict
        """
        return self._description

    def get_user(self):
        """
        User detail (partial)

        :rtype: :class:`patchwork.User.User`
        """
        return User(self._connection, self._user)



"""
PyPatchwork - Python Library for Patchwork REST API

Copyright (C) 2023  Tedd Ho-Jeong An <hj.tedd.an@gmail.com>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/
"""
import patchwork.People


class Comment:
    """
    Class for Comment object
    """

    def __init__(self, connection, attributes):
        self._connection = connection
        self._id = None
        self._web_url = None
        self._msgid = None
        self._list_archive_url = None
        self._date = None
        self._subject = None
        self._submitter = None
        self._headers = None
        self._content = None
        self.__update_attributes(attributes)

    def __update_attributes(self, attributes):
        if "id" in attributes:
            self._id = attributes["id"]
        if "web_url" in attributes:
            self._web_url = attributes["web_url"]
        if "msgid" in attributes:
            self._msgid = attributes["msgid"]
        if "list_archive_url" in attributes:
            self._list_archive_url = attributes["list_archive_url"]
        if "date" in attributes:
            self._date = attributes["date"]
        if "subject" in attributes:
            self._subject = attributes["subject"]
        if "submitter" in attributes:
            self._submitter = attributes["submitter"]
        if "headers" in attributes:
            self._headers = attributes["headers"]
        if "content" in attributes:
            self._content = attributes["content"]

    @property
    def id(self):
        """
        :type: integer
        """
        return self._id

    @property
    def web_url(self):
        """
        :type: string
        """
        return self._web_url

    @property
    def msgid(self):
        """
        :type: string
        """
        return self._msgid

    @property
    def list_archive_url(self):
        """
        :type: string
        """
        return self._list_archive_url

    @property
    def date(self):
        """
        :type: string
        """
        return self._date

    @property
    def subject(self):
        """
        :type: string
        """
        return self._subject

    @property
    def headers(self):
        """
        :type: dict
        """
        return self._headers

    @property
    def content(self):
        """
        :type: string
        """
        return self._content

    def get_submitter(self):
        """
        Submitter detail (partial)

        :rtype: :class:`patchwork.People.People`
        """
        return patchwork.People.People(self._connection, self._submitter)



"""
PyPatchwork - Python Library for Patchwork REST API

Copyright (C) 2023  Tedd Ho-Jeong An <hj.tedd.an@gmail.com>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/
"""
import patchwork.User


class People:
    """
    Class for People object
    """

    def __init__(self, connection, attributes):
        self._connection = connection
        self._id = None
        self._url = None
        self._name = None
        self._email = None
        self._user = None
        # Update attributes
        self._update_attributes(attributes)

    def _update_attributes(self, attributes):
        if "id" in attributes:
            self._id = attributes["id"]
        if "url" in attributes:
            self._url = attributes["url"]
        if "name" in attributes:
            self._name = attributes["name"]
        if "email" in attributes:
            self._email = attributes["email"]
        if "user" in attributes:
            self._user = attributes["user"]

    def __repr__(self):
        return f"People(id:{self._id} name:{self._name} email:{self._email})"

    @property
    def id(self):
        """
        :type: integer
        """
        return self._id

    @property
    def url(self):
        """
        :type: string
        """
        return self._url

    @property
    def name(self):
        """
        :type: string
        """
        return self._name

    @property
    def email(self):
        """
        :type: string
        """
        return self._email

    def get_user_info(self):
        """
        User info if exist

        :rtype: :class:`patchwork.User.User`
        """
        if self._user is None:
            return None
        return patchwork.User.User(self._connection, self._user)



"""
PyPatchwork - Python Library for Patchwork REST API

Copyright (C) 2023  Tedd Ho-Jeong An <hj.tedd.an@gmail.com>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/
"""
import os
import unittest
import patchwork


class TestBundle(unittest.TestCase):
    def setUp(self):
        self.assertIn(
            "PW_USER_TOKEN", os.environ, "Missing PW_TOKEN environment variable"
        )
        self.pw = patchwork.Patchwork(
            "https://patchwork.kernel.org", os.environ["PW_USER_TOKEN"]
        )
        self.bundle = self.pw.get_bundle(34802)

    def testNoToken(self):
        self.pw = patchwork.Patchwork("https://patchwork.kernel.org")
        with self.assertRaises(patchwork.NotFoundException):
            self.bundle = self.pw.get_bundle(34802)

    def testBasicAttributes(self):
        self.assertEqual(self.bundle.id, 34802)
        self.assertEqual(
            self.bundle.url, "https://patchwork.kernel.org/api/bundles/34802/"
        )
        self.assertEqual(
            self.bundle.web_url,
            "https://patchwork.kernel.org/bundle/tedd_an/Fix%20unchecked%20return%20value/",
        )
        self.assertEqual(self.bundle.name, "Fix unchecked return value")
        self.assertFalse(self.bundle.public)
        self.assertEqual(
            self.bundle.mbox,
            "https://patchwork.kernel.org/bundle/tedd_an/Fix%20unchecked%20return%20value/mbox/",
        )

    def testProject(self):
        project = self.bundle.get_project()
        self.assertEqual(project.id, 395)
        self.assertEqual(project.url, "https://patchwork.kernel.org/api/projects/395/")
        self.assertEqual(project.name, "Bluetooth")
        self.assertEqual(project.link_name, "bluetooth")
        self.assertEqual(project.list_id, "linux-bluetooth.vger.kernel.org")
        self.assertEqual(project.list_email, "linux-bluetooth@vger.kernel.org")
        self.assertEqual(project.web_url, "")
        self.assertEqual(project.scm_url, "")
        self.assertEqual(project.webscm_url, "")
        self.assertEqual(project.list_archive_url, "")
        self.assertEqual(
            project.list_archive_url_format, "https://lore.kernel.org/r/{}"
        )
        self.assertEqual(project.commit_url_format, "")

    def testOwner(self):
        user = self.bundle.get_owner()
        self.assertIsInstance(user, patchwork.User.User)
        self.assertEqual(user.id, 104215)
        self.assertEqual(user.url, "https://patchwork.kernel.org/api/users/104215/")
        self.assertEqual(user.username, "tedd_an")
        self.assertEqual(user.first_name, "Tedd")
        self.assertEqual(user.last_name, "An")
        self.assertEqual(user.email, "tedd.an@intel.com")

    def testPatches(self):
        patches = self.bundle.get_patches()
        self.assertIsNotNone(patches)
        self.assertEqual(len(patches), 9)
        patch_0 = patches[0]
        self.assertIsInstance(patch_0, patchwork.Patch.Patch)
        self.assertEqual(patch_0.id, 12567487)
        self.assertEqual(
            patch_0.url, "https://patchwork.kernel.org/api/patches/12567487/"
        )
        self.assertEqual(
            patch_0.web_url,
            "https://patchwork.kernel.org/project/bluetooth/patch/20211018172833.534191-2-hj.tedd.an@gmail.com/",
        )
        self.assertEqual(
            patch_0.msgid, "<20211018172833.534191-2-hj.tedd.an@gmail.com>"
        )
        self.assertEqual(
            patch_0.list_archive_url,
            "https://lore.kernel.org/r/20211018172833.534191-2-hj.tedd.an@gmail.com",
        )
        self.assertEqual(patch_0.date, "2021-10-18T17:28:25")
        self.assertEqual(patch_0.name, "[BlueZ,1/9] device: Fix unchecked return value")
        self.assertEqual(
            patch_0.mbox,
            "https://patchwork.kernel.org/project/bluetooth/patch/20211018172833.534191-2-hj.tedd.an@gmail.com/mbox/",
        )



"""
PyPatchwork - Python Library for Patchwork REST API

Copyright (C) 2023  Tedd Ho-Jeong An <hj.tedd.an@gmail.com>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/
"""
import unittest
import patchwork


class TestComment(unittest.TestCase):
    def setUp(self):
        self.pw = patchwork.Patchwork("https://patchwork.kernel.org")
        self.patch = self.pw.get_patch(12567487)
        comments = self.patch.get_comments()
        self.assertIsNotNone(comments)
        self.assertTrue(len(comments) >= 1)
        self.comment = comments[0]
        self.assertIsInstance(self.comment, patchwork.Comment.Comment)

    def testBasicAttributes(self):
        self.assertEqual(self.comment.id, 24531077)
        self.assertEqual(
            self.comment.web_url, "https://patchwork.kernel.org/comment/24531077/"
        )
        self.assertEqual(
            self.comment.msgid, "<616db65f.1c69fb81.c4a83.fe17@mx.google.com>"
        )
        self.assertEqual(
            self.comment.list_archive_url,
            "https://lore.kernel.org/r/616db65f.1c69fb81.c4a83.fe17@mx.google.com",
        )
        self.assertEqual(self.comment.date, "2021-10-18T18:01:03")
        self.assertEqual(self.comment.subject, "RE: Fix unchecked return value")
        self.assertTrue(self.comment.content.startswith("This is automated email"))

    def testHeaders(self):
        headers = self.comment.headers
        # Only check a few of them..
        self.assertIn("From", headers)
        self.assertEqual(headers["From"], "bluez.test.bot@gmail.com")
        self.assertIn("Subject", headers)
        self.assertEqual(headers["Subject"], "RE: Fix unchecked return value")

    def testSubmitter(self):
        submitter = self.comment.get_submitter()
        self.assertEqual(submitter.id, 191843)
        self.assertEqual(
            submitter.url, "https://patchwork.kernel.org/api/people/191843/"
        )
        self.assertEqual(submitter.name, None)
        self.assertEqual(submitter.email, "bluez.test.bot@gmail.com")



"""
PyPatchwork - Python Library for Patchwork REST API

Copyright (C) 2023  Tedd Ho-Jeong An <hj.tedd.an@gmail.com>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/
"""
import os
import unittest
import patchwork


class TestPatchwork(unittest.TestCase):
    def setUp(self):
        self.assertIn(
            "PW_USER_TOKEN", os.environ, "Missing PW_TOKEN environment variable"
        )
        self.pw = patchwork.Patchwork(
            "https://patchwork.kernel.org", os.environ["PW_USER_TOKEN"]
        )

    def testInvalidParameters(self):
        with self.assertRaises(AssertionError):
            patchwork.Patchwork(123)
        with self.assertRaises(AssertionError):
            patchwork.Patchwork("dummy", 123)

    def testGet_CheckInstance(self):
        self.assertIsInstance(self.pw.get_user(104215), patchwork.User.User)
        self.assertIsInstance(self.pw.get_project(395), patchwork.Project.Project)
        self.assertIsInstance(self.pw.get_people(196023), patchwork.People.People)
        self.assertIsInstance(self.pw.get_series(565705), patchwork.Series.Series)
        self.assertIsInstance(self.pw.get_cover_letter(12567485), patchwork.Cover.Cover)
        self.assertIsInstance(self.pw.get_patch(12567487), patchwork.Patch.Patch)
        self.assertIsInstance(self.pw.get_bundle(34802), patchwork.Bundle.Bundle)

    def testGet_InvalidParameter(self):
        with self.assertRaises(AssertionError):
            self.pw.get_user("104215")
        with self.assertRaises(AssertionError):
            self.pw.get_project("395")
        with self.assertRaises(AssertionError):
            self.pw.get_people("196023"),
        with self.assertRaises(AssertionError):
            self.pw.get_series("565705")
        with self.assertRaises(AssertionError):
            self.pw.get_cover_letter("12567485")
        with self.assertRaises(AssertionError):
            self.pw.get_patch("12567487")
        with self.assertRaises(AssertionError):
            self.pw.get_bundle("34802")

    def runGetAll(self, func, cls, max_count=35):
        all = func()
        count = 0
        for item in all:
            if count >= max_count:
                break
            self.assertIsInstance(item, cls)
            count += 1

    def testGetAll(self):
        self.runGetAll(self.pw.get_all_projects, patchwork.Project.Project)
        self.runGetAll(self.pw.get_all_users, patchwork.User.User)
        self.runGetAll(self.pw.get_all_people, patchwork.People.People)
        self.runGetAll(self.pw.get_all_series, patchwork.Series.Series)
        self.runGetAll(self.pw.get_all_cover_letters, patchwork.Cover.Cover)
        self.runGetAll(self.pw.get_all_patches, patchwork.Patch.Patch)
        self.runGetAll(self.pw.get_all_bundles, patchwork.Bundle.Bundle)



"""
PyPatchwork - Python Library for Patchwork REST API

Copyright (C) 2023  Tedd Ho-Jeong An <hj.tedd.an@gmail.com>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/
"""
import os
import unittest
import patchwork


class TestPagination(unittest.TestCase):
    def setUp(self):
        self.assertIn(
            "PW_USER_TOKEN", os.environ, "Missing PW_TOKEN environment variable"
        )
        self.pw = patchwork.Patchwork(
            "https://patchwork.kernel.org", os.environ["PW_USER_TOKEN"]
        )

    def testIteration(self):
        projects = self.pw.get_all_projects()
        for project in projects:
            self.assertIsInstance(project, patchwork.Project.Project)
        self.assertEqual(len(list(projects)), 100)



"""
PyPatchwork - Python Library for Patchwork REST API

Copyright (C) 2023  Tedd Ho-Jeong An <hj.tedd.an@gmail.com>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/
"""
import unittest
import patchwork


class TestProject(unittest.TestCase):
    def setUp(self):
        self.pw = patchwork.Patchwork("https://patchwork.kernel.org")
        self.bt = self.pw.get_project(395)

    def testBasicAttributes(self):
        self.assertEqual(self.bt.id, 395)
        self.assertEqual(self.bt.url, "https://patchwork.kernel.org/api/projects/395/")
        self.assertEqual(self.bt.name, "Bluetooth")
        self.assertEqual(self.bt.link_name, "bluetooth")
        self.assertEqual(self.bt.list_id, "linux-bluetooth.vger.kernel.org")
        self.assertEqual(self.bt.list_email, "linux-bluetooth@vger.kernel.org")
        self.assertEqual(self.bt.web_url, "")
        self.assertEqual(self.bt.scm_url, "")
        self.assertEqual(self.bt.webscm_url, "")
        self.assertEqual(self.bt.subject_match, "")
        self.assertEqual(self.bt.list_archive_url, "")
        self.assertEqual(
            self.bt.list_archive_url_format, "https://lore.kernel.org/r/{}"
        )
        self.assertEqual(self.bt.commit_url_format, "")

    def testMaintainersList(self):
        maintainers = self.bt.get_maintainers()
        self.assertEqual(len(maintainers), 6)
        for maintainer in maintainers:
            self.assertIsInstance(maintainer, patchwork.User.User)



"""
PyPatchwork - Python Library for Patchwork REST API

Copyright (C) 2023  Tedd Ho-Jeong An <hj.tedd.an@gmail.com>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/
"""


class Pagination:
    def __init__(
        self,
        cls,
        connection,
        next_link,
        next_params=None,
    ):
        self.__elements = list()
        self.__cls = cls
        self.__connection = connection
        self.__next_link = next_link
        self.__next_params = next_params

    def __iter__(self):
        yield from self.__elements
        while self.__next_link is not None:
            # Get next page
            new_element = self._get_next_page()
            self.__elements += new_element
            yield from new_element

    def _get_next_link(self, headers):
        links = {}
        if "Link" in headers:
            link_list = headers["Link"].split(", ")
            for link in link_list:
                url, rel = link.split("; ")
                url = url[1:-1]
                rel = rel[5:-1]
                links[rel] = url
        return links

    def _get_next_page(self):
        print(f"link: {self.__next_link}")
        headers, data = self.__connection.request(
            "GET",
            self.__next_link,
        )
        data = data if data else []

        self.__next_link = None
        if len(data) > 0:
            links = self._get_next_link(headers)
            if "next" in links:
                self.__next_link = links["next"]
        self.__next_params = None

        content = []
        for item in data:
            content.append(self.__cls(self.__connection, item))

        return content



"""
PyPatchwork - Python Library for Patchwork REST API

Copyright (C) 2023  Tedd Ho-Jeong An <hj.tedd.an@gmail.com>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/
"""
import patchwork.User


class Project:
    """
    Class for project object
    """

    def __init__(self, connection, attributes):
        self._connection = connection
        self._id = None
        self._url = None
        self._name = None
        self._link_name = None
        self._list_id = None
        self._list_email = None
        self._web_url = None
        self._scm_url = None
        self._webscm_rul = None
        self._maintainers = None
        self._subject_match = None
        self._list_archive_url = None
        self._list_archive_url_format = None
        self._commit_url_format = None
        self.__update_attributes(attributes)

    def __update_attributes(self, attributes):
        if "id" in attributes:
            self._id = attributes["id"]
        if "url" in attributes:
            self._url = attributes["url"]
        if "name" in attributes:
            self._name = attributes["name"]
        if "link_name" in attributes:
            self._link_name = attributes["link_name"]
        if "list_id" in attributes:
            self._list_id = attributes["list_id"]
        if "list_email" in attributes:
            self._list_email = attributes["list_email"]
        if "web_url" in attributes:
            self._web_url = attributes["web_url"]
        if "scm_url" in attributes:
            self._scm_url = attributes["scm_url"]
        if "webscm_url" in attributes:
            self._webscm_url = attributes["webscm_url"]
        if "maintainers" in attributes:
            self._maintainers = attributes["maintainers"]
        if "subject_match" in attributes:
            self._subject_match = attributes["subject_match"]
        if "list_archive_url" in attributes:
            self._list_archive_url = attributes["list_archive_url"]
        if "list_archive_url_format" in attributes:
            self._list_archive_url_format = attributes["list_archive_url_format"]
        if "commit_url_format" in attributes:
            self._commit_url_format = attributes["commit_url_format"]

    def __repr__(self):
        return f"Project(id:{self._id} name:{self._name})"

    @property
    def id(self):
        """
        :type: integer
        """
        return self._id

    @property
    def url(self):
        """
        :type: string
        """
        return self._url

    @property
    def name(self):
        """
        :type: string
        """
        return self._name

    @property
    def link_name(self):
        """
        :type: string
        """
        return self._link_name

    @property
    def list_id(self):
        """
        :type: string
        """
        return self._list_id

    @property
    def list_email(self):
        """
        :type: string
        """
        return self._list_email

    @property
    def web_url(self):
        """
        :type: string
        """
        return self._web_url

    @property
    def scm_url(self):
        """
        :type: string
        """
        return self._scm_url

    @property
    def webscm_url(self):
        """
        :type: string
        """
        return self._webscm_url

    @property
    def subject_match(self):
        """
        :type: string
        """
        return self._subject_match

    @property
    def list_archive_url(self):
        """
        :type: string
        """
        return self._list_archive_url

    @property
    def list_archive_url_format(self):
        """
        :type: string
        """
        return self._list_archive_url_format

    @property
    def commit_url_format(self):
        """
        :type: string
        """
        return self._commit_url_format

    def get_maintainers(self):
        """
        List of maintainers

        :rtype: List of :class:`patchwork.User.User`
        """
        maintainers = []
        for maintainer in self._maintainers:
            maintainers.append(patchwork.User.User(self._connection, maintainer))
        return maintainers



"""
PyPatchwork - Python Library for Patchwork REST API

Copyright (C) 2023  Tedd Ho-Jeong An <hj.tedd.an@gmail.com>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/
"""
import unittest
import patchwork


class TestCheck(unittest.TestCase):
    def setUp(self):
        self.pw = patchwork.Patchwork("https://patchwork.kernel.org")
        self.patch = self.pw.get_patch(12567487)
        checks = self.patch.get_checks()
        self.assertIsNotNone(checks)
        self.assertTrue(len(checks) >= 1)
        self.check = checks[0]
        self.assertIsInstance(self.check, patchwork.Check.Check)

    def testBasicAttributes(self):
        self.assertEqual(self.check.id, 745895)
        self.assertEqual(
            self.check.url,
            "https://patchwork.kernel.org/api/patches/12567487/checks/745895/",
        )
        self.assertEqual(self.check.date, "2021-10-18T17:47:10.203518")
        self.assertEqual(self.check.state, "success")
        self.assertEqual(
            self.check.target_url, "https://github.com/BluezTestBot/bluez/pull/1042"
        )
        self.assertEqual(self.check.context, "checkpatch")
        self.assertEqual(self.check.description, "Checkpatch PASS")

    def testUser(self):
        user = self.check.get_user()
        self.assertIsInstance(user, patchwork.User.User)
        self.assertEqual(user.id, 104215)
        self.assertEqual(user.url, "https://patchwork.kernel.org/api/users/104215/")
        self.assertEqual(user.username, "tedd_an")
        self.assertEqual(user.first_name, "Tedd")
        self.assertEqual(user.last_name, "An")
        self.assertEqual(user.email, "tedd.an@intel.com")


"""
"""
